/****************************************************************
ConfigServer.C
CS 481 - Team E
Spring 1998
Hummingbird Project - Dr. Frincke
Coded by: Jamie Marconi
Description: This is the implementation of the configuration server.
             The server waits for clients to connect and then 
	     executes their SQL commands.
*****************************************************************/
/*
 Copyright (C) 1997-1998 The HummingBird Project - hummer@cs.uidaho.edu

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

static char ConfigServer_C[] = "$Id: ConfigServer.C,v 1.12 1998/03/09 04:37:19 marc9442 Exp $";

#include "ConfigServer.h"

/**********
Function: ConfigServer::ConfigServer
Purpose:  Basic constructor.
**********/

ConfigServer::ConfigServer()
{
}


/**********
Function: ConfigServer::ConfigServer(NewPort)
Purpose:  Constructor that takes a port to run on.
**********/

ConfigServer::ConfigServer(int NewPort)
{
  Port = NewPort;
}


/**********
Function: ConfigServer::~ConfigServer
Purpose:  Basic destructor.
**********/

ConfigServer::~ConfigServer()
{
}


/**********
Function: ConfigServer::StartServer
Purpose:  This is where the main server loop is. This server forks off
          processes, so it can handle multiple connections at once.
**********/

void ConfigServer::StartServer(void)
{
  int Flag;
  
  // Create the sockets and initialize.
  InSocket = new sock_serv();
  InSocket->create();

  // Bind the socket to the CONFIG port, die if failure
  if (!InSocket->bind(Port, CONFIG_SERVICE)) {
    Print_Error("ConServ","Socket Bind Error. Is there a hummer already running on this machine?");
    exit(1);
  }

  // Write the PID to DB
  // Connect to DB server
  PgDatabase *piddb = OpenDB();
  
  // Delete any remained pids from DB
  QueryDB(&piddb,"delete from pid where name = 'ConServ';");

  // Create query to insert our PID
  char *query;
  query = new char[NORMAL_STRING_SIZE];
  sprintf(query, "insert into pid values ('ConServ', %ld);", getpid()); 
  
  // run query
  QueryDB(&piddb, query);

  // close DB connection
  delete [] query;
  delete piddb;

  // Listen for connections on the socket port
  InSocket->listen(SERVER_LISTEN_BACKLOG);
 
  // Repeat until SIGINT is given
  ServerRun = 1;
  while (ServerRun) { 

    // wait to accept socket connections pass of to new socket file descriptor
    if(InSocket->accept()){
      
      // fork to process connection
      if(fork() == 0) {
	// This is the child.    

	// Create a new client object to handle connection
	ConnectedSocket = InSocket->make_sock_cli();
	
	// Authenticate remote side, negotiate kerberos, die if failure
	if (!ConnectedSocket->recv_auth(CONFIG_KEYTAB)) {
	  ConnectedSocket->close();
	  delete ConnectedSocket;
	  exit(1);
	}

	// Close the listening socket, not needed in child
	InSocket->close();
	delete InSocket;
	
	// Process request in method
	ProcRequest();

	// once complete, or errored out, close and delete connection, exit child
	ConnectedSocket->close();
	delete ConnectedSocket;
	exit(0);
      } 
      else{
	// This is the Parent

	// Close the accept connection, not needed in parent.
	InSocket->close_a();
      }
    }
  }

      
  // Delete pid from database before we leave.
  piddb = OpenDB();
  QueryDB(&piddb,"delete from pid where name = 'ConServ';");
  delete piddb;

  // Once we are about to die, close out the listening socket
  InSocket->close();
  delete InSocket;
  delete ConnectedSocket;
}


/**********
Function: ConfigServer::KillServer
Purpose:  This function sets ServerRun to zero so the server exits
**********/

void ConfigServer::KillServer(void)
{
  ServerRun = 0;
}


/**********
Function: ConfigServer::ProcRequest
Purpose:  This function Processes the Requests from the Manager ConfigClient
**********/

int ConfigServer::ProcRequest(void)
{
  // Create memory area for reading stuff off the socket and sending
  // to the DB
  char *query;
  query = new char[8192];
  
  // Connect to the PostgreSQL database backend, die if failure
  PgDatabase *db = OpenDB();
  if (!db) {
    delete db;
    delete [] query;
    
    return 0;
  }
  
  // Write out string to inform the client of your service
  ConnectedSocket->write("Configuration Server Ready"); 
 
  // Do until the end of protocol indicator
  do {
    //Clear out the buffer first
    bzero(query,8192);
    
    // Read the query from the client
    ConnectedSocket->read(query, 8192);
  
#ifdef DEBUG_QUERY
    Print_Error("ConServ", query);
#endif

    // If it starts with an 'I' then we think its an SQL command
    // so run the query
    if ((query[0] == 'I') || (query[0] == 'D')) {
      QueryDB(&db, query);
    }
    
    // If its starts with a 'G' then we are told to restart the running Server
    else if (query[0] == 'G') {
      // query the DB to find the pid of the Server
      if (QueryDB(&db,"SELECT * FROM pid WHERE name = 'Server';")) {
	// Send HUP signal to the Server
	kill(atoi(db->GetValue(0,1)), SIGHUP); 
      }
    }
    
    // If the string was anything else quit.
    else {
      break;
    }

    // If the query started with an 'E' stop
  } while (query[0] != 'E');
  
  // Close connection to database backend, free memory
  delete db;
  delete [] query;
  
  return 1;
}
