/****************************************************************
PeerIMain.C
CS 481 - Team A
Spring 1997
Hummingbird Project - Dr. Frincke
Coded by: James Cooper, Jamie Marconi, based on code written by Pete Nielson
Description: This file acts as a client within peer moderation.
  It is used to connect to PeerServ in order to subscribe, unsubscribe, 
  and send messages to the Peer Server (or moderator) running somewhere
  on the internet.
*****************************************************************/
/*
 Copyright (C) 1997-1998 The HummingBird Project - hummer@cs.uidaho.edu

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

static char PeerIMain_C[] = "$Id: PeerIMain.C,v 1.18 1998/05/05 20:07:15 coop9342 Exp $";

#include "Hummer.h"         // Global hummer defines.
#include "Hummer_PG.h"
#include "Message.h"
#include "sock_cli.h"
#include "Print_Error.h"    // Print_Error fuction.
#include <iostream.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <ctype.h>


// Function prototypes
void PrintInstruct(void);
int ParseParameters(int ArgCount, char **Args);
char *GetMod(const char *peer_grp);
int CheckDateTime(char *Date);
int SubToMod();
int UnSubToMod();
int MessToMod();

// global vars for peer communication
char message[NORMAL_STRING_SIZE];
char PgName[NORMAL_STRING_SIZE];
char *ModAddr = NULL;
int SubMod = 0;
int UnSubMod = 0;
int MessMod = 0;
int Mess = 0;

/**********
Function: Main
Purpose:  This is the main function of the peer interface. It handles
          the input passed into it via command line parameters. 
**********/

int main(int argc, char **argv)
{
  
  if(argc != 1){
    // Parse the command line.
    if(ParseParameters(argc, argv)){
      PrintInstruct();
      return 0;
    }
    if ((SubMod) && (!ModAddr)) {
      PrintInstruct();
      return 0;
    }
    if (!ModAddr) {
      ModAddr = GetMod(PgName);
      if (!ModAddr) {
	Print_Error("PeerI", "Could not get Moderator host address for Peer Group %s", PgName);
      }
    }
    // Send the info to the Moderator
    if (((SubMod) || (UnSubMod)) || (MessMod)) {
      if (((SubMod) && (!UnSubMod)) && (!MessMod)) {
        // if the user wants to subscribe, call subscribe
	SubToMod();
	return 0;
      }
      else if (((UnSubMod) && (!SubMod)) && (!MessMod)) {
	// if the user wants to unsubscribe, call unsubscribe
	UnSubToMod();
	return 0;
      }
      else if ((MessMod) && (Mess) && (!SubMod) && (!UnSubMod)) {
        // if the user wants to send a message, do that
	MessToMod();
	return 0;
      }
      else {
	PrintInstruct();
	return 0;
      }
    }
    else {
      PrintInstruct();
      return 0;
    }
 }
  else{
    PrintInstruct();        // Print the instructions 
    return 0;
  }

}


/**********
Function: PrintInstruct
Purpose:  Print the instructions on how to call PeerI to stdout.
**********/

void PrintInstruct(void)
{
  // Dump the instructions
  cerr << "usage: PeerI -m <peer group> -s <peer group> -u <peer group> -M <message> -a <moderator address>" << endl;
}


/**********
Function: ParseParameters
Purpose:  Parse the command line parameters and place the the data in the 
message object. This function also reports errors in the command
line so the instructions can be printed.
**********/

int ParseParameters(int ArgCount, char **Args)
{
  int c;
  
  while ((c = getopt (ArgCount, Args, "m:s:u:M:a:")) != -1) {
#ifdef DEBUG
    //Print_Error("PeerI", "optarg addr: %p", &optarg);
#endif
    switch (c)
      {
      case 'm':
	strncpy(PgName,(char*) optarg, NORMAL_STRING_SIZE-1);
	PgName[NORMAL_STRING_SIZE-1] = '\0';
        MessMod = 1;
	break;
      case 's':
	strncpy(PgName,(char*) optarg, NORMAL_STRING_SIZE-1);
	PgName[NORMAL_STRING_SIZE-1] = '\0';
        SubMod = 1;
	break;
      case 'u':
	strncpy(PgName,(char*) optarg, NORMAL_STRING_SIZE-1);
	PgName[NORMAL_STRING_SIZE-1] = '\0';
        UnSubMod = 1;
	break;
      case 'M':
	strncpy(message,(char*) optarg, NORMAL_STRING_SIZE-1);
	message[NORMAL_STRING_SIZE-1] ='\0';
        Mess = 1;
	break;
      case 'a':
	ModAddr = new char[strlen(optarg)+1];
	strcpy(ModAddr,(char*) optarg);
	break;
      case '?':
	if (isprint (optopt))
	  cerr << "Unknown option `-" << (char) optopt << "'." << endl;
	else
	  cerr << "Unknown option character `\\x" << hex << optopt << "'." << endl;
	return 1;
      default:
	return 1;
      }
  }
  
  return 0;
}

/**********
Function: GetMod
Purpose:  Make sure the command line arguments are within some requirements
**********/

char *GetMod(const char *peer_grp)
{
  char *query;
  query = new char[8192];
  
  PgDatabase *db = OpenDB();
  if (!db) {
    cerr << "PeerI: cannot connect to database" << endl;
    delete db;
    delete [] query;
    return NULL;
  }
  sprintf(query,"SELECT * FROM pgroups WHERE pgroup = '%s';",peer_grp);
  
  if (!QueryDB(&db, query)) {
    cerr << "PeerI: cannot connect to database" << endl;
    delete db;
    delete [] query;
    return NULL; 
  }
  char *addr = new char[db->GetLength(0,"mod_addr")+1];
  strcpy(addr,db->GetValue(0,"mod_addr"));  // get mod address from database
  delete db;
  delete [] query;

  return addr;
}


/**********
Function: CheckDateTime
Purpose:  This function checks to make sure the date and time are valid. 
          At this point this just checks to make sure they are not zero.
**********/

int CheckDateTime(char *Date)
{
  int TempDate;

  TempDate = atoi(Date);
  if(TempDate<1) return 1;
  else return 0;
}


/**********
Function: SubToMod
Purpose:  Opens a connection to the Peer Server (PeerServ) and sends a request
          for subscription to the server.  Upon success, the server will send
	  database information to this client including all of the items relevant
	  to the peer group we are seeking to subscribe to.  The subscription will
	  update the tables "peers" and "pgroups" in our database.
**********/

int SubToMod()
{
  // DECLARATIONS 
  char *InMessage[NORMAL_STRING_SIZE];           // Used to receive messages
  char *ServDir;                                 // Used to receive a 'yes' or 'no' from server
  sock_cli *HummerSocket;                        // Socket Object
  int PeerTuples = 0;                            // Used to update database

  // INITIALIZATIONS
  ServDir = new char[5];
  bzero(InMessage,NORMAL_STRING_SIZE);           // Initalize InMessage
  HummerSocket = new sock_cli();                 // Set and create our new Socket Object
  HummerSocket->create();                  

  // CREATE SOCKET CONNECTION
  // If we connect to the server, call send_auth...if that doesn't work, exit
  if(HummerSocket->connect(ModAddr, PEER_PORT, PEER_SERVICE)) {
    if (!HummerSocket->send_auth()) {
      Print_Error("PeerI", "Send_Auth Failure");
      HummerSocket->close();
      delete HummerSocket;
      return 0;
    }
    // Connection successful, continue...

    // TELL SERVER WE WAHT TO SUBSCRIBE BY SENDING "s" AND PEER GROUP NAME
    HummerSocket->write("s");
    HummerSocket->write(PgName);
    // Server returns 'y' or 'n', letting us know if we have successfully subscribed or not
    HummerSocket->read((char*)ServDir,5);
    // Server tells us why or why not...
    HummerSocket->read((char*)InMessage,NORMAL_STRING_SIZE);
    Print_Error("PeerI","Connected To Server: %s, %s",ServDir,InMessage);
    // If we cannot subscribe, terminate the connection
    if (!strcmp(ServDir,"n")) {
      Print_Error("PeerI", "Failed to connect to %s", PgName);
      HummerSocket->close();
      delete HummerSocket;
      return 0;
    }
    // We can subscribe, so we continue...
    
    // UPDATE OUR DATABASE TABLES "peers","levels", AND "pgroups" WITH PEER GROUP INFO FROM SERVER
    // Read in the number of table entries for peers in our peer group
    HummerSocket->read((char*)InMessage,NORMAL_STRING_SIZE);
    PeerTuples = atoi((char*)InMessage);
    // Open our database for insertion...if it works continue, if not stop
    PgDatabase *peersdb = OpenDB();
    if (!peersdb) {
      Print_Error("PeerI","Unsuccessful attempt at opening table.");
      delete peersdb;
      delete HummerSocket;
      return 0;
    }
    // Receive all of the server's "peers" and "levels" table info regarding our peer group, and insert
    // it into our "peers" and "levels" table (all levels default to 1)
    for (int i=0;i<PeerTuples;i++) {
      HummerSocket->read((char*)InMessage,NORMAL_STRING_SIZE);   
      //die if unsuccessful, else continue...
      if (!QueryDB(&peersdb,(char*) InMessage)) {
    	Print_Error("PeerI","Table insert query unsuccessful: %s",InMessage);
    	delete peersdb;
    	delete HummerSocket;
    	return 0;
      }
      Print_Error("PeerI","Table insert query successful.");
    }
    // Receive all of the server's "pgroups" table info regarding our peer group, and 
    // insert it into our "pgroups" table
    HummerSocket->read((char*)InMessage,NORMAL_STRING_SIZE);   
    //die if unsuccessful, else continue
    if (!QueryDB(&peersdb,(char*) InMessage)) {
      Print_Error("PeerI","Table insert query unsuccessful: %s",InMessage);
      delete peersdb;
      delete HummerSocket;
      return 0;
    }
    Print_Error("PeerI","Table insert query successful.");
    // done, so let's quit...
    delete peersdb;
    HummerSocket->close();
    delete HummerSocket;
    return 1;
  }
  // Couldn't connect to Server, quit...
  else{
    Print_Error("PeerI", "Failed to connect to %s", ModAddr);
    HummerSocket->close();
    delete HummerSocket;
    return 0;
  }
}

/**********
Function: UnSubToMod
Purpose:  Opens a connection to the Peer Server (PeerServ) and sends a request
          for unsubscription to the server.  Upon success, the server will remove us
	  from their database and we will proceed to clean up our database.
**********/

int UnSubToMod()
{
  // DECLARATIONS 
  char *InMessage[NORMAL_STRING_SIZE];           // Used to receive messages
  char *ServDir;                                 // Used to receive a 'yes' or 'no' from server
  sock_cli *HummerSocket;                        // Socket Object

  // INITIALIZATIONS
  ServDir = new char[5];
  bzero(InMessage,NORMAL_STRING_SIZE);           // Initalize InMessage
  HummerSocket = new sock_cli();                 // Set and create our new Socket Object
  HummerSocket->create();                  

  // CREATE SOCKET CONNECTION
  // If we connect to the server, call send_auth...if that doesn't work, exit
  if(HummerSocket->connect(ModAddr, PEER_PORT, PEER_SERVICE)) {
    if (!HummerSocket->send_auth()) {
      Print_Error("PeerI", "Send_Auth Failure");
      HummerSocket->close();
      delete HummerSocket;
      return 0;
    }
    // Connection successful, continue...

    // TELL SERVER WE WAHT TO UNSUBSCRIBE BY SENDING "u" AND PEER GROUP NAME
    HummerSocket->write("u");
    HummerSocket->write(PgName);
    // Server returns 'y' or 'n', letting us know if we have successfully unsubscribed or not
    HummerSocket->read(ServDir,5);
    // Server tells us why or why not...
    HummerSocket->read((char*)InMessage,NORMAL_STRING_SIZE);
    Print_Error("PeerI","Connected To Server: %s, %s",ServDir,InMessage);
    // If we cannot unsubscribe, terminate the connection
    if (!strcmp(ServDir,"n")) {
      Print_Error("PeerI", "Failed to connect to %s", PgName);
      HummerSocket->close();
      delete HummerSocket;
      return 0;
    }
    // We have unsubscribed, so we continue...

    // CLEAN UP OUR DATABASE (REMOVE OUR SUBSCRIPTION TO PgName)
    // Create a Database Object, if successful perform deletions...
    PgDatabase *peer = OpenDB();
    if (peer) {
      char* query = new char[8192];
      sprintf(query, "delete from peers where pgroup = '%s';", PgName);
      QueryDB(&peer, query);
      sprintf(query, "delete from pgroups where pgroup = '%s';", PgName);
      QueryDB(&peer, query);
      delete [] query;
      Print_Error("PeerI","Successfully removed from moderator's database.");
    }
    // Object creation unsuccessful, quit!
    else {
      Print_Error("PeerI","Error removing peer from moderator's database.");
      HummerSocket->close();
      delete HummerSocket;
      return 0;
    }
    // Everything looks good, quit!
    HummerSocket->close();
    delete HummerSocket;
    return 1;
  }
  // Connection to Server bad, bad, very bad...quit!
  else{
    Print_Error("PeerI", "Failed to connect to %s", ModAddr);
    HummerSocket->close();
    delete HummerSocket;
    return 0;
  }
}

/**********
Function: MessToMod
Purpose:  Opens a connection to the Peer Server (PeerServ) and sends a request
          for sending a message to the peer group via the server. 
**********/


int MessToMod()
{
 

  // DECLARATIONS 
  Message *PeerMessage;                          // Message object
  char *TempString;                              // To help with time calcs
  tm *TimeStruct;                                // Time structure returns current time
  time_t *TempTime;                              
  TempTime = new time_t;
  char *OutMessage;                              // For sending info to server
  char InMessage[NORMAL_STRING_SIZE];            // For receiving info from server
  char ServDir[5];                               // Used to receive a "y" or "n" from the server
  sock_cli *HummerSocket;                        // Socket Object                          

  // INITIALIZATIONS
  bzero(InMessage,NORMAL_STRING_SIZE);
  PeerMessage = new Message;
  TempString = new char[255];
  HummerSocket = new sock_cli();
  HummerSocket->create();

  // CREATE SOCKET CONNECTION
  // If we connect to the server, call send_auth...if that doesn't work, exit
  if(HummerSocket->connect(ModAddr, PEER_PORT, PEER_SERVICE)) {
    if (!HummerSocket->send_auth()) {
      Print_Error("PeerI", "Send_Auth Failure");
      HummerSocket->close();
      delete HummerSocket;
      return 0;
    }
    // Connection successful, continue...

    // TELL SERVER WE WAHT TO SEND A MESSAGE BY SENDING "m" AND PEER GROUP NAME
    HummerSocket->write("m");
    HummerSocket->write(PgName);
    // Server returns 'y' or 'n', letting us know if we can send a message...
    HummerSocket->read(ServDir,5);
    // Server tells us why or why not...
    HummerSocket->read(InMessage,NORMAL_STRING_SIZE);
    Print_Error("PeerI","Connected To Server: %s, %s",ServDir,InMessage);
    // If we cannot send a message, terminate the connection
    if (!strcmp(ServDir,"n")) {
      Print_Error("PeerI", "Failed to connect to %s", PgName);
      HummerSocket->close();
      delete HummerSocket;
      return 0;
    }
    // We can send a message, so we continue...

    // FORMAT AND SEND THE MESSAGE
    // Find the time (not necessarily y2k compliant :o)
    time(TempTime);
    TimeStruct = localtime(TempTime);
    sprintf(TempString, "%.4d%.2d%.2d", TimeStruct->tm_year+1900, TimeStruct->tm_mon+1, TimeStruct->tm_mday); 
    strcpy(PeerMessage->Date, TempString);
    sprintf(TempString, "%.2d%.2d%.2d", TimeStruct->tm_hour, TimeStruct->tm_min, TimeStruct->tm_sec); 
    
    // Set up struct to send to Server
    strcpy(PeerMessage->Time, TempString);
    strcpy(PeerMessage->Description,message);
    strcpy(PeerMessage->Filename,PgName);
    strcpy(PeerMessage->Tool,"peers");
    
    // Get the local host name
    if (gethostname(PeerMessage->HostIP, IP_ADDRESS_SIZE) < 0) {
      return 0;
    }
    
    // Send struct to Server
    HummerSocket->write(PeerMessage->Description);
    HummerSocket->write(PeerMessage->Date);
    HummerSocket->write(PeerMessage->Time);
    HummerSocket->write(PeerMessage->Filename);
    HummerSocket->write(PeerMessage->Tool);
    HummerSocket->write(PeerMessage->HostIP);
    HummerSocket->write("0");
    HummerSocket->close();
    delete[] message;
    delete HummerSocket;
    return 1;
    
  }
  // Connection Failed, Quit!
  else{
    Print_Error("PeerI", "Failed to connect to %s", ModAddr);
    HummerSocket->close();
    delete HummerSocket;
    return 0;
  } 
}












