/****************************************************************
PeerServer.C
CS 481 - Team E
Spring 1998
Hummingbird Project - Dr. Frincke
Coded by: James Cooper, Jamie Marconi
Description: This is the implementation of the peer moderation server.
             The server waits for peers to connect (subscribe), grants
	     or denies a subscription, and moves to a message handling
	     state if appropriate.
*****************************************************************/
/*
 Copyright (C) 1997-1998 The HummingBird Project - hummer@cs.uidaho.edu

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

static char PeerServer_C[] = "$Id: PeerServer.C,v 1.18 1998/04/03 22:47:49 coop9342 Exp $";


#include "PeerServer.h"
 


/**********
Function: PeerServer::PeerServer
Purpose:  Basic constructor.
**********/

PeerServer::PeerServer()
{
}


/**********
Function: PeerServer::PeerServer(NewPort)
Purpose:  Constructor that takes a port to run on.
**********/

PeerServer::PeerServer(int NewPort)
{
  Port = NewPort;
}


/**********
Function: PeerServer::~PeerServer
Purpose:  Basic destructor.
**********/

PeerServer::~PeerServer()
{
}


/**********
Function: PeerServer::StartServer
Purpose:  This is where the main server loop is. This server forks off
          processes, so it can handle multiple connections at once.
**********/

void PeerServer::StartServer(void)
{
  int Flag;
  
  char PeerAddr[IP_ADDRESS_SIZE];  
  // Create the sockets and initialize.
  InSocket = new sock_serv();
  InSocket->create();

  // Bind the socket to the CONFIG port, die if failure
  if (!InSocket->bind(Port, PEER_SERVICE)) {
    fprintf(stderr,"Socket Bind Error. Is there a hummer already running on this machine?\n");
    exit(1);
  }

  // Write the PID to DB
  // Connect to DB server
  PgDatabase *piddb = OpenDB();
  
  // Delete any remained pids from DB
  QueryDB(&piddb,"delete from pid where name = 'PeerServ';");

  // Create query to insert our PID
  char *query;
  query = new char[NORMAL_STRING_SIZE];
  sprintf(query, "insert into pid values ('PeerServ', %ld);", getpid()); 
  
  // run query
  QueryDB(&piddb, query);

  // close DB connection
  delete [] query;
  delete piddb;

  // Listen for connections on the socket port
  InSocket->listen(SERVER_LISTEN_BACKLOG);
 
  // Repeat until SIGINT is given
  ServerRun = 1;
  PushToPeers();

  while (ServerRun) { 

    // wait to accept socket connections pass of to new socket file descriptor
    if(InSocket->accept()){
      
      // fork to process connection
      if(fork() == 0) {
	// This is the child.    

	// Create a new client object to handle connection
	ConnectedSocket = InSocket->make_sock_cli();
	
	// Authenticate remote side, negotiate kerberos, die if failure
	if (!ConnectedSocket->recv_auth(CONFIG_KEYTAB)) {
	  ConnectedSocket->close();
	  delete ConnectedSocket;
	  exit(1);
	}

	// Get the address of the client before we close the socket.
	InSocket->get_rmt_addr(PeerAddr);

	// Close the listening socket, not needed in child
	InSocket->close();

	delete InSocket;
	
	// Process request in method
	ProcRequest(PeerAddr);

	// once complete, or errored out, close and delete connection, exit child
	ConnectedSocket->close();
	delete ConnectedSocket;
	exit(0);
      } 
      else{
	// This is the Parent

	// Close the accept connection, not needed in parent.
	InSocket->close_a();
      }
    }
    // Has the alarm go off to send config to peers.
    if (PushPeerInfo) {
      // Call Push Peer Info method
      PushToPeers();
    }
  }

  // Delete pid from database before we leave.
  piddb = OpenDB();
  QueryDB(&piddb,"delete from pid where name = 'PeerServ';");
  delete piddb;

  // Once we are about to die, close out the listening socket
  InSocket->close();
  delete InSocket;
  delete ConnectedSocket;
}


int PeerServer::PushToPeers(void)
{
  //Reset boolean
  PushPeerInfo = 0;  

  Print_Error("PeerServer::PushToPeers", "Alarm went off");
}

void PeerServer::SetPushPeer(void)
{
  PushPeerInfo = 1;
}


/**********
Function: PeerServer::KillServer
Purpose:  This function sets ServerRun to zero so the server exits
**********/

void PeerServer::KillServer(void)
{
  ServerRun = 0;
}


/**********
Function: PeerServer::ProcRequest
Purpose:  This function receives a connection from PeerI, a peer
          seeking to subscribe, and grants/denies a subscription
	  based on table info in the "peers" table.
**********/

int PeerServer::ProcRequest(char *PeerAddr)
{

  // Create memory area for reading stuff off the socket and sending
  // to the DB
  char *query;
  char *message;
  char *PgName;
  char *PeerDir;
  int Success;
  
  query = new char[8192];
  message = new char[1000]; 
  PgName = new char[1000]; 
  PeerDir = new char[5];
  
  // Clear out the buffer first
  bzero(query,8192);
  bzero(message,1000);
  bzero(PgName,1000);

  // Read in directive from peer ("I want to sub, unsub, send message, etc...")
  ConnectedSocket->read(PeerDir, 5);
  Print_Error("PeerServer::ProcessRequest", "Directive from Peer: %s", PeerDir);

    
  // Read in a group name from a peer
  ConnectedSocket->read(PgName, 1000);
  Print_Error("PeerServer::ProcessReques","Read Peer Group Name: %s",PgName);
  
  // Open a connection to the DB, die if unsuccessful
  PgDatabase *peerdb = OpenDB();
  if (!peerdb) {
    delete peerdb;
    delete [] query;
    delete [] message;
    delete [] PgName;
    return 0;
  }   
  
  // See if peer group name AND peer address are in our table
  sprintf(query, "SELECT * FROM peers WHERE pgroup = '%s' AND mem_addr = '%s';\n"
	  ,PgName,PeerAddr); // Peer Address from get_rmt_address
    
  // Perform the query, quit and send message to peer if peer isn't in our table
  if (!QueryDB(&peerdb, query)) {
    Print_Error("PeerServer::ProcessReques","Subscriber \"%s\", Peer Group \"%s\" NOT listed in peer table."
		,PeerAddr,PgName);
    sprintf(query,"\"%s\" in group \"%s\" is not a valid combination."
	    ,PeerAddr,PgName);
    ConnectedSocket->write("n");
    ConnectedSocket->write(query);
    delete peerdb;
    delete [] query;
    delete [] message;
    delete [] PgName;    
    return 0; 
  }
    
  // Peer group name exists in table with peer address, continue connection and update peer
  Print_Error("PeerServer::ProcessReques","Peer Group is: %s",peerdb->GetValue(0,0));
  Print_Error("PeerServer::ProcessReques","Member Address is: %s",peerdb->GetValue(0,1));
  ConnectedSocket->write("y");
  delete peerdb;

  // if peer wishes to subscribe, do this
  if (!strcmp(PeerDir,"s")) {
    Print_Error("PeerServer::ProcessReques","Peer Directive is subscribe.");
    ConnectedSocket->write("Doing subscribe work...");
    Success = PeerSubscribe(PeerAddr,PgName,query);
  }

  // if peer wishes to unsubscribe, do this
  else if (!strcmp(PeerDir,"u")) {
    Print_Error("PeerServer::ProcessReques","Peer Directive is unsubscribe.");
    ConnectedSocket->write("Doing unsubscribe work...");
    Success = PeerUnSubscribe(PeerAddr,PgName,query);
  }

  // if peer wishes to send a message, do this
  else if (!strcmp(PeerDir,"m")) {
    Print_Error("PeerServer::ProcessReques","Peer Directive is to send a message.");
    ConnectedSocket->write("Doing message work...");
    Success = PeerMessage(PeerAddr,PgName,query,message);
  }

  else {
    Print_Error("PeerServer::ProcessReques", "PeerI directive unmatched: '%s'", PeerDir);
    delete [] query;
    delete [] message;
    delete [] PgName;
    delete [] PeerAddr;
    return 0;
  }
  
  delete [] query;
  delete [] message;
  delete [] PgName;
  delete [] PeerAddr;
  return 1;
}


/**********
Function: PeerServer::PeerSubscribe
Purpose:  This function allows PeerI to subscribe to the peergroup
**********/

int PeerServer::PeerSubscribe(char *PeerAddr, char *PgName, char *query)
{
  // Open a connection to the DB, die if unsuccessful
  PgDatabase *peerdb = OpenDB();
  if (!peerdb) {
    delete peerdb;
    delete [] query;
    delete [] PeerAddr;
    delete [] PgName;
    return 0;
  }   
  
  // Update our table to indicate that this peer is subscribed to our group
  sprintf(query, 
	  "UPDATE peers SET peer_sub = 1, peer_mod=0 WHERE pgroup = '%s' AND mem_addr ='%s';"
	  ,PgName,PeerAddr); 
  // Perform the update, quit if update doesn't go through
  if (!QueryDB(&peerdb, query)) {
    Print_Error("PeerServer::PeerSubscribe","Table update query unsuccessful.");
    delete peerdb;
    delete [] query;
    delete [] PeerAddr;
    delete [] PgName;
    return 0;
  }
  // Update successful, continue connection
  Print_Error("PeerServer::PeerSubscribe","Table update query successful.");  
  // Get the peers with our subscriber's group in common, and send them to the peer
  sprintf(query,"SELECT * FROM peers WHERE pgroup='%s';",PgName); 
  // Perform the query, quit if select doesn't go through
  if (!QueryDB(&peerdb, query)) {
    delete peerdb;
    delete [] query;
    delete [] PeerAddr;
    delete [] PgName;
    return 0;
  }
  
  // Tell the peer how many times items it must read into its database in order to 
  // add the peers from the moderator's table
  sprintf(query,"%d",peerdb->Tuples());
  ConnectedSocket->write(query);
  
  // For every row returned create an SQL statement that will insert
  // the row on the server's side DB (for the peers and levels tables)
  for (int i=0; i < peerdb->Tuples(); i++) {   
    sprintf(query, "INSERT INTO peers VALUES ( '%s', '%s', %s, %s);INSERT INTO levels VALUES ('%s','',1,1,1,0);\n",
	    peerdb->GetValue(i,0),  // pgroup
	    peerdb->GetValue(i,1),  // mem_addr
	    peerdb->GetValue(i,2),  // peer_sub
	    peerdb->GetValue(i,3),  // peer_mod
            peerdb->GetValue(i,1)); // mem_addr
    ConnectedSocket->write(query);
  }
  
 sprintf(query,"SELECT * FROM pgroups WHERE pgroup='%s';",PgName); 
  // Perform the query, quit if select doesn't go through
  if (!QueryDB(&peerdb, query)) {
    delete peerdb;
    delete [] query;
    delete [] PeerAddr;
    delete [] PgName;
    return 0;
  }
  sprintf(query, "INSERT INTO pgroups VALUES ( '%s', '%s', '%s');DELETE FROM levels WHERE start_addr='%s';\n",
	  peerdb->GetValue(0,0),   // pgroup
	  peerdb->GetValue(0,1),   // mod_addr
	  peerdb->GetValue(0,2),   // email
          PeerAddr);               // peer address
  ConnectedSocket->write(query);
  // close DB connection
  delete peerdb;
  delete [] query;
  delete [] PeerAddr;
  delete [] PgName;
  return 1;
}


/**********
Function: PeerServer::PeerUnsubscribe
Purpose:  This function processes a request from PeerI to unsubscribe from moderator
**********/
int PeerServer::PeerUnSubscribe(char *PeerAddr, char *PgName, char *query)
{
  // Open a connection to the DB, die if unsuccessful
  PgDatabase *peerdb = OpenDB();
  if (!peerdb) {
    delete peerdb;
    delete [] query;
    delete [] PeerAddr;
    delete [] PgName;
    return 0;
  }   
  
  sprintf(query,"Delete FROM peers Where pgroup = '%s' AND mem_addr = '%s';"
	  ,PgName,PeerAddr);;
  // Perform the query, quit and send message to peer if deletion is not successful
  if (!QueryDB(&peerdb, query)) {
    Print_Error("PeerServer::PeerUnsubscribe","Could not delete values.",PeerAddr,PgName);
    ConnectedSocket->write("n");
    delete peerdb;
    delete [] query;
    delete [] PeerAddr;
    delete [] PgName;
    return 0;
  }
  Print_Error("PeerServer::PeerUnsubscribe","Peer successfully deleted from table peers.");
  ConnectedSocket->write("y");

  // close DB connection
  delete peerdb;
  delete [] query;
  delete [] PeerAddr;
  delete [] PgName;
  return 1;
}



/**********
Function: PeerServer::PeerMessage
Purpose:  This function processes a message request from PeerI
**********/
int PeerServer::PeerMessage(char *PeerAddr, char *PgName, char *query, char *message)
{
  // Allocate space to hold data
  Message *PeerMessage;
  char *InMessage;
  char *OutMessage;
  PeerMessage = new Message;

  InMessage = new char[5];

  Print_Error("\nPeerServer::PeerMessage","Hummer Message will be:");

  // Receive message from PeerI, to be sent to all peers in group
  ConnectedSocket->read(PeerMessage->Description,NORMAL_STRING_SIZE);
  ConnectedSocket->read(PeerMessage->Date,DATE_SIZE);
  ConnectedSocket->read(PeerMessage->Time,TIME_SIZE);
  ConnectedSocket->read(PeerMessage->Filename,NORMAL_STRING_SIZE);
  ConnectedSocket->read(PeerMessage->Tool,NORMAL_STRING_SIZE);
  ConnectedSocket->read(PeerMessage->HostIP,IP_ADDRESS_SIZE);

  // Message aging still a problem this is not an ideal world here.
  // Someday we might want to use a Message Digest on Hummer Messages
  // and keep a list to check against.
  ConnectedSocket->read(InMessage,5);
  PeerMessage->MessageAge = max(atoi(InMessage), MAX_MESSAGE_AGE-2);

  // Report to Server Message attributes
  Print_Error("PeerServer::PeerMessage","Hummer Message will be:\nDate: %s\nTime: %s\nMessage: %s\nPeer Group: %s\nPeer Address: %s\nTool: %s\nAge: %d",
	      PeerMessage->Date,PeerMessage->Time,PeerMessage->Description,
	      PeerMessage->Filename,PeerMessage->HostIP,PeerMessage->Tool,
	      PeerMessage->MessageAge);
  
  // Open a connection to the DB, die if unsuccessful
  PgDatabase *peerdb = OpenDB();
  if (!peerdb) {
    delete peerdb;
    delete [] query;
    delete [] PeerAddr;
    delete [] PgName;
    return 0;
  }   
  
  // Get all addresses of subscribed hosts, perform query, die if unsuccessful
  sprintf(query,"SELECT mem_addr FROM peers WHERE pgroup = '%s' and peer_sub = 1;",PgName);
  if (!QueryDB(&peerdb, query)) {
    delete peerdb;
    delete [] query;
    delete [] PeerAddr;
    delete [] PgName;
    return 0;
  }
  
  OutMessage = new char[BUFFER_SIZE];
  // Set up the outgoing message.
  sprintf(OutMessage, "%d|%s|%s|%s|%s|%s|%s",
	  PeerMessage->MessageAge,                      // Keep forwarding?
	  PeerMessage->Date,                            // Date
	  PeerMessage->Time,                            // Time
	  PeerMessage->Description,                     // Message
	  PeerMessage->HostIP,                          // Author's IP
	  PeerMessage->Filename,                        // Source IP
	  PeerMessage->Tool);                           // Source Tool
  
  // For each peer address returned, connect to it and send the message
  for (int i=0; i < peerdb->Tuples(); i++) {   
    sock_cli *HummerSocket;
    
    HummerSocket = new sock_cli();
    HummerSocket->create();
    
    strcpy(query,peerdb->GetValue(i,0));
    Print_Error("PeerServer::PeerMessage", "Connecting to %s", query);
    // If we get a connection, do the following
    if(HummerSocket->connect(query, MESSAGE_PORT, MESSAGE_SERVICE)) {
      
      
      // Send the message, die if failed
      if (!HummerSocket->send_auth()) {
	Print_Error("PeerServer::PeerMessage", "Send_Auth Failure");
	HummerSocket->close();
      }
      
      // Send the message
      Print_Error("PeerServer::PeerMessage", "Sent to peer: %s", OutMessage);
      HummerSocket->write(OutMessage);
      HummerSocket->close();  
    }
    // No Connection Established, quit
    else{
      Print_Error("PeerServer::PeerMessage", "Failed to connect to %s"
		  ,peerdb->GetValue(i,0));
      HummerSocket->close();
    }        
    delete HummerSocket;
  }
  delete peerdb;
  
  // close DB connection
  delete [] query;
  delete [] PeerAddr;
  delete [] PgName;
  return 1;
}












































