/******************************************************************
krb5_sock_cli.C
HummingBird Project - Deborah A. Frincke, PhD.
Summer 1997
Coded by: Jamie Marconi
Description: This class is used for kreberized client connections.  It
is also used to hadn off server connection to a socket after a connection
has been authenicated.
*******************************************************************/
/*
  Copyright (C) 1997-1998 The HummingBird Project - hummer@cs.uidaho.edu
 
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <config.h>

#ifdef KERBEROS

#include "krb5_sock_cli.h"

/*******
Function: krb5_sock_cli::krb5_sock_cli
Purpose:  Basic constructor. A few variable inits.
*******/

krb5_sock_cli::krb5_sock_cli()
{
  keytab_name = NULL;
  principal_name = NULL;
}


/*******
Function: krb5_sock_cli::krb5_sock_cli
Purpose:  Constructor for passed auth-info and open socket from server.
          Auth-info is not copied, authorization must occur after this call.
*******/

krb5_sock_cli::krb5_sock_cli(int asock, char * pservice)
{
  service = strdup(pservice);
  sock = asock;

  retval = krb5_init_context(&context);
  if (retval) {
    com_err("krb5_sock::krb5_create()", retval, "while initializing krb5");
  }
}

/*******
Function: krb5_sock_cli::krb5_sock_cli
Purpose:  Basic destructor
********/

krb5_sock_cli::~krb5_sock_cli()
{
  if (auth_context) 
    krb5_auth_con_free(context, auth_context);

  auth_context = NULL;
}

/*********
Fucntion: krb5_sock_cli::krb5_set_names
Purpose:  fill in the client keytab, and principal names
*********/

krb5_sock_cli::krb5_set_names(char *keytab, char *principal)
{
  keytab_name = keytab;
  principal_name = principal;
}

/**********
Function: krb5_sock_cli::krb5_connect
Purpose:  Connects to the hostname and port specified with pservice.
          Fills in all host information, local and remote.
          Returns 0 on error, 1 otherwise
**********/

int krb5_sock_cli::krb5_connect(char *remote_host, int port, char* pservice)
{
  if (!pservice) {
    com_err("krb5_sock_cli::krb5_connect", 0, "invalid service name");
    return 0;
  } 

  service = pservice;
 
  // Make sure we got a remote hostname
  if (!remote_host) {
    com_err("krb5_sock_cli::krb5_connect", 0, "invalid remote host");
    return 0;
  }

  // Look up fully qualified remote hostname
  struct hostent *host;
  if ((host = gethostbyname(remote_host)) == (struct hostent *) 0) {
    fprintf(stderr, "%s: unknown host\n", remote_host);
    return 0;
  }

  strncpy(full_rmt_hname, host->h_name, sizeof(full_rmt_hname)-1);
  full_rmt_hname[sizeof(full_rmt_hname)-1] = '\0';

  for (char *cp = full_rmt_hname; *cp; cp++)
    if (isupper(*cp))
      *cp = tolower(*cp);
  
  // Fill out a sockaddr_in for the remote hostname
  //  struct sockaddr_in rmt_sockaddr;
  (void) memset((char *) rmt_sockaddr, 0, sizeof(struct sockaddr_in));
  memcpy((char *) &rmt_sockaddr->sin_addr, host->h_addr, sizeof(rmt_sockaddr->sin_addr));

  // and the port and type
  rmt_sockaddr->sin_family = AF_INET;
  rmt_sockaddr->sin_port = htons(port);
  
#ifdef DEBUG
  fprintf(stderr, "rmt_sockaddr.sin_addr is %s\n", inet_ntoa(rmt_sockaddr->sin_addr));
  fprintf(stderr, "rmt_sockaddr.sin_port is %d\n", ntohs(rmt_sockaddr->sin_port));
#endif

  if (connect(sock, (struct sockaddr *)rmt_sockaddr, sizeof(struct sockaddr_in)) < 0) {
    perror("connect");
    close(sock);
    return 0;
  }

  return 1;
}


/**********
Function: krb5_sock_cli::krb5_connect
Purpose:  Connects to the hostname  specified with pservice.
          Fills in all host information, local and remote.
          Returns 0 on error, 1 otherwise
**********/

int krb5_sock_cli::krb5_connect(char *remote_host, char* pservice)
{
  struct servent *serv;

  if (!pservice) {
    com_err("krb5_sock_cli::krb5_connect", 0, "invalid service name");
    return 0;
  } 

  service = pservice;

  if ((serv = getservbyname(pservice, "tcp")) == NULL) {
    fprintf(stderr, "service unknown: %s/tcp\n", pservice);
    return 0;
  }

  return krb5_connect(remote_host, ntohs(serv->s_port), pservice);
}

/**********
Function: krb5_sock_cli::krb5_s_auth
Purpose:  Send authenication material to server and authenicate remote host.
          returns 0 on error, 1 on success
**********/ 

int krb5_sock_cli::krb5_s_auth()
{
  // Fill out a sockaddr_in for the local hostname
  //struct sockaddr_in loc_sockaddr;
  memset((char *)loc_sockaddr, 0, sizeof(struct sockaddr_in));
  loc_sockaddr->sin_family = AF_INET;
#ifdef BROKEN_STREAMS_SOCKETS
  if (gethostname(full_loc_hname, sizeof(full_loc_hname)) < 0) {
    perror("gethostname");
    return 0;
  }
  
  struct hostent *host;
  if ((host = gethostbyname(full_loc_hname)) == (struct hostent *)0) {
    fprintf(stderr, "%s: unknown host\n", full_loc_hname);
    return 0;
  }
  memcpy((char *) &loc_sockaddr->sin_addr, host->h_addr, sizeof(loc_sockaddr->sin_addr));
#endif /* BROKEN_STREAMS_SOCKETS */ 

  int sockaddr_in_len = sizeof(struct sockaddr_in);
  if (getsockname(sock, (struct sockaddr *) loc_sockaddr, &sockaddr_in_len) < 0) {
    perror("getsockname");
    close(sock);
    return 0;
  }

#ifdef DEBUG
  fprintf(stderr, "loc_sockaddr.sin_addr is %s\n", inet_ntoa(loc_sockaddr->sin_addr));
  fprintf(stderr, "loc_sockaddr.sin_port is %d\n", ntohs(loc_sockaddr->sin_port));
#endif

  krb5_ccache ccache = NULL;
  krb5_principal client;
  krb5_principal server;

  if (retval = krb5_cc_default(context, &ccache)) {
    com_err("krb5_sock_cli::krb5_m_auth", retval, "while getting default ccache");
    return 0;
  }

  if (!keytab_name && !principal_name) {
    if (retval = krb5_cc_get_principal(context, ccache, &client)) {
      com_err("krb5_sock_cli::krb5_m_auth", retval, "while getting client principal name");
      return 0;
    }
  }
  else {

    if (retval = krb5_cc_get_principal(context, ccache, &client)) {

      krb5_data tgtname = {
	0,
	KRB5_TGS_NAME_SIZE,
	KRB5_TGS_NAME
      };
      
      krb5_preauthtype * preauth = NULL;
      int options = KRB5_DEFAULT_OPTIONS;
      int option;
      int errflg = 0;
      krb5_creds my_creds;
      krb5_address *null_addr = (krb5_address *)0;
      krb5_address **addrs = (krb5_address **)0;
      krb5_keytab keytab = NULL;
      char  *client_name;
      
      retval = krb5_kt_resolve(context, keytab_name, &keytab);
      if (retval != 0) {
	com_err("kinit", retval, "resolving keytab %s",
		keytab_name);
	errflg++;
      }
      
      if ((retval = krb5_parse_name (context, principal_name, &client))) {
	com_err ("kinit", retval, "when parsing name %s",principal_name);
	return 0;
      }
      
      if ((retval = krb5_unparse_name(context, client, &client_name))) {
	com_err ("kinit", retval, "when unparsing name");
	return 0;
      }
      
      memset((char *)&my_creds, 0, sizeof(my_creds));
      
      my_creds.client = client;
      
      if((retval = krb5_build_principal_ext(context, &server,
					    krb5_princ_realm(context, client)->length,
					    krb5_princ_realm(context, client)->data,
					    tgtname.length, tgtname.data,
					    krb5_princ_realm(context, client)->length,
					    krb5_princ_realm(context, client)->data,
					    0))) {
	com_err("kinit", retval, "while building server name");
	return 0;
      }
      
      my_creds.server = server;
      
      retval = krb5_get_in_tkt_with_keytab(context, options, addrs,
					   NULL, preauth, keytab, 0,
					   &my_creds, 0);
      
      if (retval) {
	if (retval == KRB5KRB_AP_ERR_BAD_INTEGRITY)
	  fprintf (stderr, "%s: Password incorrect\n", "kinit");
	else
	  com_err ("kinit", retval, "while getting initial credentials");
	return 0;
      }
      
      retval = krb5_cc_initialize (context, ccache, client);
      if (retval != 0) {
	com_err ("kinit", retval, "when initializing cache ");
	return 0;
      }
      
      retval = krb5_cc_store_cred(context, ccache, &my_creds);
      if (retval) {
	com_err ("kinit", retval, "while storing credentials");
	return 0;
      }
    }
    else {
#ifdef DEBUG 
      com_err("krb5_sock_cli::krb5_m_auth", 0, "Using Default Cache name");
#endif
    }
  }
  
  if (retval = krb5_sname_to_principal(context, full_rmt_hname, service,
				       KRB5_NT_SRV_HST, &server)) {
    com_err("krb5_sock_cli::krb5_m_auth", retval, "while creating server name for %s/%s",
	    full_rmt_hname, service);
    return 0;
  }
  
  krb5_data cksum_data;
  cksum_data.data = full_rmt_hname;
  cksum_data.length = strlen(full_rmt_hname);

  krb5_error *err_ret;
  krb5_ap_rep_enc_part *rep_ret;
  retval = krb5_sendauth(context, &auth_context, (krb5_pointer) &sock,
			 KRB5_SOCK_VERSION, client, server,
			 AP_OPTS_MUTUAL_REQUIRED,
			 &cksum_data,
			 0,		/* no creds, use ccache instead */
			 ccache, &err_ret, &rep_ret, NULL);
  
  krb5_free_principal(context, server);	/* finished using it */ 
  krb5_free_principal(context, client);
  
  if (retval && retval != KRB5_SENDAUTH_REJECTED) {
    com_err("krb5_sock_cli::krb5_m_auth", retval, "while using sendauth");
    return 0;
  }
  if (retval == KRB5_SENDAUTH_REJECTED) {
    /* got an error */
#ifdef DEBUG
    fprintf(stderr, "sendauth rejected, error reply is:\n\t\"%*s\"\n",
	   err_ret->text.length, err_ret->text.data);
#endif
    return 0;
  } else if (rep_ret) {
    /* got a reply */
#ifdef DEBUG
    fprintf(stderr, "sendauth succeeded\n");
#endif
  } else {
    com_err("krb5_sock_cli::krb5_m_auth", 0, "no error or reply from sendauth!");
    return 0;
  }

  krb5_address loc_addr;
  krb5_address rmt_addr;

  loc_addr.addrtype = ADDRTYPE_IPPORT;
  loc_addr.length = sizeof(loc_sockaddr->sin_port);
  loc_addr.contents = (krb5_octet *) &loc_sockaddr->sin_port;

  rmt_addr.addrtype = ADDRTYPE_IPPORT;
  rmt_addr.length = sizeof(rmt_sockaddr->sin_port);
  rmt_addr.contents = (krb5_octet *) &rmt_sockaddr->sin_port;

  if ((retval = krb5_auth_con_setports(context, auth_context,
				       &loc_addr, &rmt_addr))) {
    com_err("krb5_sock_cli::krb5_m_auth", retval, "while setting local port\n");
    return 0;
  }

  loc_addr.addrtype = ADDRTYPE_INET;
  loc_addr.length = sizeof(loc_sockaddr->sin_addr);
  loc_addr.contents = (krb5_octet *) &loc_sockaddr->sin_addr;
  
  rmt_addr.addrtype = ADDRTYPE_INET;
  rmt_addr.length = sizeof(rmt_sockaddr->sin_addr);
  rmt_addr.contents = (krb5_octet *) &rmt_sockaddr->sin_addr;

  if ((retval = krb5_auth_con_setaddrs(context, auth_context,
				       &loc_addr, &rmt_addr))) {
    com_err("krb5_sock_cli::krb5_m_auth", retval, "while setting local addr\n");
    return 0;
  }

  krb5_address *portlocal_addr;
  if ((retval = krb5_gen_portaddr(context, &loc_addr,
				  (krb5_pointer) &loc_sockaddr->sin_port,
				  &portlocal_addr))) {
    com_err("krb5_sock_cli::krb5_m_auth", retval, "while generating port address");
    return 0;
  }

  char *cp;
  if ((retval = krb5_gen_replay_name(context, portlocal_addr,
				     REPLAY_UNIQUE_NAME, &cp))) {
    com_err("krb5_sock_cli::krb5_m_auth", retval, "while generating replay cache name");
    return 0;
  }

  krb5_data rcache_name;
  rcache_name.length = strlen(cp);
  rcache_name.data = cp;

  krb5_rcache rcache;
  if ((retval = krb5_get_server_rcache(context, &rcache_name, &rcache))) {
    com_err("krb5_sock_cli::krb5_m_auth", retval, "while getting server rcache");
    return 0;
  }

  /* set auth_context rcache */
  krb5_auth_con_setrcache(context, auth_context, rcache);
  
  krb5_xfree(cp);

  return 1;
}

/********
Function: krb5_sock_serv::krb5_r_auth
Purpose: Recieve and process Kerberos Authorization information
********/

char * krb5_sock_cli::krb5_r_auth(char *keytab_name)
{
  if (retval = krb5_kt_resolve(context, keytab_name, &keytab)) {
    com_err("krb5_sock_serv::krb5_m_auth", retval,"while resolving keytab file %s", keytab_name);
    return NULL;
  }

  krb5_principal server;
  if (retval = krb5_sname_to_principal(context, NULL, service, 
				       KRB5_NT_SRV_HST, &server)) {
    com_err("krb5_sock_serv::krb5_m_auth", retval,
	    "while generating service name (%s)", service);
    return NULL;
  }

  if (getpeername(sock, (struct sockaddr *) rmt_sockaddr, &sin_len) < 0) {
    fprintf(stderr, "getpeername error\n");
    return NULL;
  }

#ifdef DEBUG
  fprintf(stderr, "rmt_sockaddr.sin_addr is %s\n", inet_ntoa(rmt_sockaddr->sin_addr));
  fprintf(stderr, "rmt_sockaddr.sin_port is %d\n", ntohs(rmt_sockaddr->sin_port)); 
#endif
  
  krb5_ticket * ticket;
  if (retval = krb5_recvauth(context, &auth_context, (krb5_pointer)&sock,
			     KRB5_SOCK_VERSION, server, 
			     0,	/* no flags */
			     keytab,	/* default keytab is NULL */
			     &ticket)) {
    com_err("krb5_sock_serv::krb5_m_auth", retval, "recvauth failed");
    return NULL;
  }

  char *cname;
  if (retval = krb5_unparse_name(context, ticket->enc_part2->client, &cname)){
    com_err("krb5_sock_serv::krb5_m_auth", retval, "unparse failed");
#ifdef DEBUG
    fprintf(stderr, "You are <unparse error>\n");
#endif
  } else {
#ifdef DEBUG
    fprintf(stderr, "You are %s\n", cname);
#endif
    //free(cname);
  }

  // Fill out a sockaddr_in for the local hostname
  //  struct sockaddr_in loc_sockaddr;
  memset((char *)loc_sockaddr, 0, sizeof(struct sockaddr_in));
  loc_sockaddr->sin_family = AF_INET;
#ifdef BROKEN_STREAMS_SOCKETS
  if (gethostname(full_loc_hname, sizeof(full_loc_hname)) < 0) {
    perror("gethostname");
    return NULL;
  }
  
  struct hostent *host;
  if ((host = gethostbyname(full_loc_hname)) == (struct hostent *)0) {
    fprintf(stderr, "%s: unknown host\n", full_loc_hname);
    return NULL;
  }
  memcpy((char *) &loc_sockaddr->sin_addr, host->h_addr, sizeof(loc_sockaddr->sin_addr));
#endif /* BROKEN_STREAMS_SOCKETS */ 
  
  int sockaddr_in_len = sizeof(struct sockaddr_in);
  if (getsockname(sock, (struct sockaddr *) loc_sockaddr, &sockaddr_in_len) < 0) {
    perror("getsockname");
    close(sock);
    return NULL;
  }
  
#ifdef DEBUG
  fprintf(stderr, "loc_sockaddr.sin_addr is %s\n", inet_ntoa(loc_sockaddr->sin_addr));
  fprintf(stderr, "loc_sockaddr.sin_port is %d\n", ntohs(loc_sockaddr->sin_port));
#endif

  krb5_address rmt_addr;
  krb5_address loc_addr;
  
  rmt_addr.addrtype = ADDRTYPE_IPPORT;
  rmt_addr.length = sizeof(rmt_sockaddr->sin_port);
  rmt_addr.contents = (krb5_octet *) &rmt_sockaddr->sin_port;

  loc_addr.addrtype = ADDRTYPE_IPPORT;
  loc_addr.length = sizeof(loc_sockaddr->sin_port);
  loc_addr.contents = (krb5_octet *) &loc_sockaddr->sin_port;

  if ((retval = krb5_auth_con_setports(context, auth_context,
				       &loc_addr, &rmt_addr))) {
    com_err("krb5_sock_serv::krb5_m_auth", retval, "while setting remote port\n");
    return NULL;
  }

  rmt_addr.addrtype = ADDRTYPE_INET;
  rmt_addr.length = sizeof(rmt_sockaddr->sin_addr);
  rmt_addr.contents = (krb5_octet *) &rmt_sockaddr->sin_addr;

  loc_addr.addrtype = ADDRTYPE_INET;
  loc_addr.length = sizeof(loc_sockaddr->sin_addr);
  loc_addr.contents = (krb5_octet *) &loc_sockaddr->sin_addr;

  if ((retval = krb5_auth_con_setaddrs(context, auth_context,
				       &loc_addr, &rmt_addr))) {
    com_err("krb5_sock_serv::krb5_m_auth", retval, "while setting remote addr\n");
    return NULL;
  }

  return cname;
}


/**********
Function: krb5_sock_cli::krb5_e_read
Purpose:  Reads the data from the socket unencryped and source verified then,
          puts it in the buffer.
          Returns 0 on error, 1 otherwise
**********/
int krb5_sock_cli::krb5_e_read(char *buf, size_t buflen)
{
  // unsigned char pktbuf[BUFSIZ];
  krb5_data packet;
  krb5_data message;
  short xmitlen;

  if ((retval = krb5_net_read(context, sock, (char *)&xmitlen,
			      sizeof(xmitlen))) <= 0) {
    if (retval == 0)
      errno = ECONNABORTED;
    com_err("krb5_sock_cli::krb5_e_read", errno, 
	    "while reading data from server");
    return 0;
  }
  packet.length = ntohs(xmitlen);
  if (!(packet.data = (char *)malloc(packet.length + 1))) {
    com_err("krb5_sock_cli::krb5_e_read", ENOMEM,
	    "while allocating buffer to read from server");
    return 0;
  }
  if ((retval = krb5_net_read(context, sock, (char *)packet.data,
			      packet.length)) <= 0) {
    if (retval == 0)
      errno = ECONNABORTED;
    com_err("krb5_sock_cli::krb5_e_read", errno, 
	    "while reading data from server");
    return 0;
  }

#ifdef DEBUG
  fprintf(stderr, "Received %d bytes\n", packet.length);
#endif
  
  //packet.length = i;
  //packet.data = buf;//(krb5_pointer) buf;
  
  if ((retval = krb5_rd_priv(context, auth_context, &packet,
			     &message, NULL))) {
    com_err("krb5_sock_cli::krb5_e_read", retval, 
	    "while verifying PRIV message");
    return 0;
  }

  free((void *)packet.data);
  
  strcpy(buf, message.data);
  return message.length;
}


/**********
Function: krb5_sock_cli::krb5_e_write
Purpose:  Writes message to the socket encryped and source verified then,
          Returns 1 on error, 0 otherwise
**********/
int krb5_sock_cli::krb5_e_write(char *buf)
{
  krb5_data packet;
  krb5_data inbuf;
  short xmitlen;

  inbuf.data = buf;
  inbuf.length = strlen(buf);

  if ((retval = krb5_mk_priv(context, auth_context, &inbuf,
			     &packet, NULL))) {
    com_err("krb5_sock_cli::krb5_e_write", retval, 
	    "while making KRB_PRIV message");
    return 0;
  }

  xmitlen = htons(packet.length);
  if ((retval = krb5_net_write(context, sock, (char *)&xmitlen,
			       sizeof(xmitlen))) < 0) {
    com_err("krb5_sock_cli::krb5_e_write", errno, 
	    "while writing len to client");
    return 0;
    }
  if ((retval = krb5_net_write(context, sock, (char *)packet.data,
			       packet.length)) < 0) {
    com_err("krb5_sock_cli::krb5_e_write", errno,
	    "while writing data to client");
    return 0;
  }

  krb5_xfree(packet.data);
  
  return packet.length;
}








#endif /* KERBEROS */
